
    /*
     * Autosscep -- Simple SCEP client implementation
     * Copyright (c) Alberto Forino 2004. All rights reserved.
     * See the file COPYRIGHT for licensing information.
     */


#include "conf.h"
#include "cmd.h"

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <openssl/evp.h>
#include <openssl/crypto.h>
#include <openssl/buffer.h>
#include <openssl/asn1.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pkcs7.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/rand.h>
#include <openssl/md5.h>
#include <openssl/objects.h>
#include <openssl/asn1_mac.h>


/* Global defines */

/* SCEP operations */
int operation_flag;
#define	SCEP_OPERATION_GETCA	1
#define	SCEP_OPERATION_ENROLL	3
#define	SCEP_OPERATION_GETCERT	5
#define	SCEP_OPERATION_GETCRL	7
#define	SCEP_OPERATION_AUTOENROLL	9

/* SCEP MIME headers */
#define MIME_GETCA	"application/x-x509-ca-cert"
#define MIME_GETCA_RA	"application/x-x509-ca-ra-cert"
/* Entrust VPN connector uses different MIME types */
#define MIME_PKI	"x-pki-message"
#define MIME_GETCA_RA_ENTRUST	"application/x-x509-ra-ca-certs"

/* SCEP reply types based on MIME headers */
#define	SCEP_MIME_GETCA		1
#define	SCEP_MIME_GETCA_RA	3
#define	SCEP_MIME_PKI		5

/* SCEP request types */
#define	SCEP_REQUEST_NONE		0
#define	SCEP_REQUEST_PKCSREQ		19
#define	SCEP_REQUEST_PKCSREQ_STR	"19"
#define	SCEP_REQUEST_GETCERTINIT	20
#define	SCEP_REQUEST_GETCERTINIT_STR	"20"
#define	SCEP_REQUEST_GETCERT		21
#define	SCEP_REQUEST_GETCERT_STR	"21"
#define	SCEP_REQUEST_GETCRL		22
#define	SCEP_REQUEST_GETCRL_STR		"22"

/* SCEP reply types */
#define	SCEP_REPLY_NONE		0
#define	SCEP_REPLY_CERTREP	3
#define	SCEP_REPLY_CERTREP_STR	"3"

/* SCEP pkiStatus values (also used as SSCEP return values) */
#define SCEP_PKISTATUS_SUCCESS		0
#define SCEP_PKISTATUS_FAILURE		2
#define SCEP_PKISTATUS_PENDING		3

/* SSCEP return values (not in SCEP draft) */
#define SCEP_PKISTATUS_ERROR		1	/* General error */
#define SCEP_PKISTATUS_BADALG		70	/* BADALG failInfo */
#define SCEP_PKISTATUS_BADMSGCHK	71	/* BADMSGCHK failInfo */
#define SCEP_PKISTATUS_BADREQ		72	/* BADREQ failInfo */
#define SCEP_PKISTATUS_BADTIME		73	/* BADTIME failInfo */
#define SCEP_PKISTATUS_BADCERTID	74	/* BADCERTID failInfo */
#define SCEP_PKISTATUS_TIMEOUT		89	/* Network timeout */
#define SCEP_PKISTATUS_SS		91	/* Error generating selfsigned */
#define SCEP_PKISTATUS_FILE		93	/* Error in file handling */
#define SCEP_PKISTATUS_NET		95	/* Network sending message */
#define SCEP_PKISTATUS_P7		97	/* Error in pkcs7 routines */
#define SCEP_PKISTATUS_UNSET		99	/* Unset pkiStatus */

/* SCEP failInfo values */
#define SCEP_FAILINFO_BADALG		0
#define SCEP_FAILINFO_BADALG_STR	\
	"Unrecognized or unsupported algorithm ident"
#define SCEP_FAILINFO_BADMSGCHK		1
#define SCEP_FAILINFO_BADMSGCHK_STR	\
	"Integrity check failed"
#define SCEP_FAILINFO_BADREQ		2
#define SCEP_FAILINFO_BADREQ_STR	\
	"Transaction not permitted or supported"
#define SCEP_FAILINFO_BADTIME		3
#define SCEP_FAILINFO_BADTIME_STR	\
	"Message time field was not sufficiently close to the system time"
#define SCEP_FAILINFO_BADCERTID		4
#define SCEP_FAILINFO_BADCERTID_STR 	\
	"No certificate could be identified matching"

	#define DEFAULT_BITS 1024
	
	/* error (autoscep) definition */
#define	ERR_CA_BAD_ENC_ALGO -1
#define ERR_CA_CERT_MISSING -2
#define ERR_CA_ENC_CERT_MISSING -3
#define ERR_CA_BAD_CONFIG -4
#define ERR_CERT_MISSING_KEY -5
#define ERR_CERT_MISSING_CERT -6
#define ERR_CERT_BAD_CONFIG -7
#define ERR_CERT_MISSING_CA -8
#define ERR_CERT_KEY_MISSING -9
#define ERR_CERT_REQ_MISSING -10
#define ERR_CERT_BAD_REQ -11
#define ERR_CERT_CREATING_SELFSIGNED -12
#define ERR_CERT_PKCS7_WRAP -13
#define ERR_CERT_HTTP -14
#define ERR_CERT_PKCS7_UNWRAP -15
#define ERR_CERT_BAD_REPLY -16
#define ERR_IO -17
#define ERR_CA_FAILED_CACERT_REQ -18

#define TEST_FILE_NAME "autoscep_test_file"
/* End of Global defines */


/* Global variables */

/* auxiliary temporary string*/
char           temp_string[1024];
#define temp_string_len 1024

/* Program name */
char *pname;

/* Network timeout */
int timeout;

/* Fingerprint, signing and encryption algorithms */
EVP_MD *fp_alg;
EVP_CIPHER *enc_alg;
EVP_MD *sig_alg;

/* OpenSSL OID handles */
int nid_messageType;
int nid_pkiStatus;
int nid_failInfo;
int nid_senderNonce;
int nid_recipientNonce;
int nid_transId;
int nid_extensionReq;

/* Global pkistatus */
int pkistatus;

/* End of Global variables */


/* Structures */

/* GETCertInital data structure */

typedef struct
{
  X509_NAME *issuer;
  X509_NAME *subject;
} pkcs7_issuer_and_subject;

/* HTTP reply structure */
struct http_reply
{

  /* SCEP reply type */
  int type;

  /* Status */
  int status;

  /* Payload */
  char *payload;

  /* Payload size */
  int bytes;
};

/* SCEP transaction structure */
struct scep
{

  /* SCEP message types */
  int request_type;
  char *request_type_str;
  int reply_type;
  char *reply_type_str;

  /* SCEP message status */
  int pki_status;
  char *pki_status_str;
  int fail_info;
  char *fail_info_str;

  /* SCEP transaction attributes */
  char *transaction_id;
  unsigned char *sender_nonce;
  int sender_nonce_len;
  unsigned char *reply_recipient_nonce;
  unsigned char *reply_sender_nonce;
  int recipient_nonce_len;

  /* Certificates */
  X509 *selfsigned;
  EVP_PKEY *pkey;

  /* Request */
  PKCS7 *request_p7;
  unsigned char *request_payload;
  int request_len;
  pkcs7_issuer_and_subject *ias_getcertinit;
  PKCS7_ISSUER_AND_SERIAL *ias_getcert;
  PKCS7_ISSUER_AND_SERIAL *ias_getcrl;

  /* Reply */
  PKCS7 *reply_p7;
  unsigned char *reply_payload;
  int reply_len;

};

struct cert_data
{
  /* dati per il salvataggio/lettura su/da file */
  char *cert_file;		// stringa con il nome del file
  char *key_file;		//stringa con il nome del file della chiave privata

  /* strutture gestione certificati, richieste e chiavi */
  X509 *cert;			// certificato X509
  EVP_PKEY *key;
  EVP_PKEY *new_key;		// chiave privata per una nuova richiesta
  X509_REQ *req;		// richiesta di certificato
    STACK_OF (X509_EXTENSION) * exts;  // estensioni X509v3
  char *CADN; // DN della CA firmataria

  /* valori dei campi per DN */
  char *C;
  char *ST;
  char *L;
  char *O;
  char *OU;
  char *CN;
  char *Email;
  
  /*valore per l'autenticazione automatica*/
  char *challenge_password;

  /* valori per ripristinare la transazione scep */
  char *transactionID;

  /* flag e messaggio di errore */
  int err_flag;

  /* flag sullo stato del certificato */
  int renewed_flag;
  
  /* command for requesting (first time ) cert to the CA (ex: e-mail script)*/
  char *req_command;
  
};

struct ca_data
{
	/* Distinguisced Name */
  char *DN;
  /* Scep script url*/
  char *url_scep;
  /* certificate file name */
  char *cert_file;
  /* Ecnription certificate file name*/
  char *enc_file;
  /* host name */
  char *host;
  /* SCEP script path*/
  char *host_dir;
  /* host port */
  char *port;
  /* encription algorithm string*/
  char *enc_alg_str;
  /* CA identifier*/
  char *ID;
  /* encription file flag */
  int ef_flag;
  /* vitual host flag */
  int VHost_flag;

  /* certificati */
  X509 *cert;
  X509 *enc_cert;
  EVP_CIPHER *enc_alg;

  /* flag e messaggio di errore */
  int err_flag;
  char *error;

};

/* End of structures */


/* Functions */

/* Send HTTP message */
int send_msg (struct http_reply *, char *, char *, int, int);

/* Catch SIGALRM */
void catchalarm (int);

/* Get config file parameter */
char *get_string (char *);

/* Report memory error */
void error_memory (void);

/* Initialize config file */
void init_config (FILE *);

/* Initialize SCEP layer */
int init_scep (void);

/* Write other certificate file */
void write_other_cert (struct scep *);

/* Write CA files */
int write_ca_ra(struct http_reply *s, struct ca_data *ca);

/* Create new SCEP session */
int new_transaction (struct scep *, struct cert_data *c);

/* Get key fingerprint */
char *key_fingerprint (X509_REQ *);

/* PKCS#7 encode message */
int pkcs7_wrap (struct scep *s, struct cert_data *ce, struct ca_data *ca);

/* PKCS#7 decode message */
int pkcs7_unwrap (struct scep *s, struct cert_data *ce, struct ca_data *ca);

/* Add signed string attribute */
int add_attribute_string (STACK_OF (X509_ATTRIBUTE) *, int, char *);

/* Add signed octet attribute */
int add_attribute_octet (STACK_OF (X509_ATTRIBUTE) *, int, char *, int);

/* Find signed attributes */
int get_signed_attribute (STACK_OF (X509_ATTRIBUTE) *, int, int, char **);
int get_attribute (STACK_OF (X509_ATTRIBUTE) *, int, ASN1_TYPE **);

/* URL-endcode */
char *url_encode (char *, size_t);

/* check the data in configuration file */
int check_config ();

/* read data from file and set the CA structure */
int set_ca_struct (FILE * fp, struct ca_data *c);

/* read data from file and set the certificate structure */
int set_cert_struct (FILE * fp, struct cert_data *c);

/* read certificate from file */
X509 *get_cert (char *nomefile);

/* check the validity date of certificate*/
int check_date (X509 * cert, int ndays);

/* return the ca structure with the DN of issuer of cert*/
struct ca_data *get_ca_struct (X509 * cert);

/* generate a new private key, set the cert structure and save the key in [cert->key_file].new */
int gen_new_key (struct cert_data *c);

/* generate from teh structure data a new cert request (pkcs10) */
int gen_new_req (struct cert_data *c);

/* analize the scep url and set up the host data */
int set_host_data (struct ca_data *c);

/* save a certificate on a file (PEM format) */
int write_cert (X509 * x, char *nomefile);

/* analyze the scep reply and write the cert */
void write_reply_cert (struct scep *s, struct cert_data *ce);

/* read private key from file */
EVP_PKEY *get_key (char *s);

/* write the certificates contained in the http message of scep structure */
void write_local_cert (struct scep *s, struct cert_data *ce);

/* generate a selfsigned certificate*/
X509 *new_selfsigned (struct cert_data *c);

/* create a X509v3 extension stack from config file data*/
STACK_OF (X509_EXTENSION) * set_extensions (FILE * fp, STACK_OF (X509_EXTENSION) * exts);

/* add extension to the stack from nid and value*/			    
int  add_ext (STACK_OF (X509_REQUEST) * sk, int nid, char *value);

/* check if exist the pending file */
char *ctrl_pending (struct cert_data *ce);

/* write the pending file (SCEP transaction ID) */
int write_pending (struct cert_data *ce);

/* read request from file */
X509_REQ * get_req (char *nomefile);

/* update certificate on disk*/
int update(struct cert_data *cert);

/* check the basic option parameter*/
void check_general_option();

/* request from server the ca cert(s) and save it on disk*/
void request_ca_cert(struct ca_data *ca);

/* End of Functions */

/* Global structures*/
     struct ca_data *
       CAs[10];
     struct cert_data *
       certs[100];

/*dalle librerie openssl*/
     time_t
     ASN1_UTCTIME_get (const ASN1_UTCTIME * s);
